# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""FMMAX: Fourier modal method with jax.

Copyright (c) Martin F. Schubert
"""

__version__ = "v1.5.0"

# ruff: noqa: F401
from fmmax.basis import (
    Expansion,
    LatticeVectors,
    Truncation,
)
from fmmax.basis import X as _X
from fmmax.basis import Y as _Y
from fmmax.basis import (
    brillouin_zone_in_plane_wavevector,
    generate_expansion,
    min_array_shape_for_expansion,
    plane_wave_in_plane_wavevector,
    transverse_wavevectors,
    unit_cell_coordinates,
)
from fmmax.beams import shifted_rotated_fields
from fmmax.farfield import farfield_integrated_flux, farfield_profile
from fmmax.fields import (
    amplitude_poynting_flux,
    colocate_amplitudes,
    directional_poynting_flux,
    eigenmode_poynting_flux,
    fields_from_wave_amplitudes,
    fields_on_coordinates,
    fields_on_grid,
    layer_amplitudes_interior,
    layer_fields_3d,
    layer_fields_3d_on_coordinates,
    propagate_amplitude,
    stack_amplitudes_interior,
    stack_amplitudes_interior_with_source,
    stack_fields_3d,
    stack_fields_3d_auto_grid,
    stack_fields_3d_on_coordinates,
    time_average_z_poynting_flux,
)
from fmmax.fmm import (
    Formulation,
    LayerSolveResult,
    broadcast_result,
    eigensolve_anisotropic_media,
    eigensolve_general_anisotropic_media,
    eigensolve_isotropic_media,
)
from fmmax.pml import PMLParams, apply_uniaxial_pml
from fmmax.scattering import (
    ScatteringMatrix,
    append_layer,
    prepend_layer,
    redheffer_star_product,
    stack_s_matrices_interior,
    stack_s_matrix,
    stack_s_matrix_scan,
)
from fmmax.sources import (
    amplitudes_for_fields,
    amplitudes_for_source,
    dirac_delta_source,
    gaussian_source,
)
from fmmax.translate import translate_layer_solve_result
from fmmax.utils import angular_frequency_for_wavelength, interpolate_permittivity

X = _X  # basis.X
"""Unit vector pointing in the x direction."""

Y = _Y  # basis.Y
"""Unit vector pointing in the y direction."""



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions related to eigensolve with translated permittivity distributions.

Copyright (c) Martin F. Schubert
"""

import functools

import jax.numpy as jnp

from fmmax import basis, fft, fmm, fmm_matrices


def translate_layer_solve_result(
    solve_result: fmm.LayerSolveResult,
    dx: jnp.ndarray,
    dy: jnp.ndarray,
) -> fmm.LayerSolveResult:
    """Obtain solve result for layer with translated permittivity distribution.

    Rather than performing eigensolve on translated permittivity distribution, this
    function allows one to translate the solve result. This can result in significant
    compute savings for structures such as slanted gratings.

    Args:
        solve_result: The ``LayerSolveResult`` for the permittivity distribution prior
            to translation.
        dx: The translation of the permittivity distribution along the ``x`` direction.
        dy: The translation of the permittivity distribution along the ``y`` direction.

    Returns:
        The ``LayerSolveResult`` for the layer with translated permittivity.
    """
    assert dx.shape == dy.shape == ()

    _shift = functools.partial(
        _apply_shift_eigenvectors,
        expansion=solve_result.expansion,
        primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
        dx=dx,
        dy=dy,
    )

    ev_te, ev_tm = jnp.split(solve_result.eigenvectors, 2, axis=-2)
    shifted_eigenvectors = jnp.concatenate([_shift(ev_te), _shift(ev_tm)], axis=-2)

    _shift_toeplitz = functools.partial(
        _apply_shift_toeplitz,
        expansion=solve_result.expansion,
        primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
        dx=dx,
        dy=dy,
    )

    def _shift_block_toeplitz(matrix: jnp.ndarray) -> jnp.ndarray:
        m0, m1 = jnp.split(matrix, 2, axis=-2)
        m00, m01 = jnp.split(m0, 2, axis=-1)
        m10, m11 = jnp.split(m1, 2, axis=-1)
        return jnp.block(
            [
                [_shift_toeplitz(m00), _shift_toeplitz(m01)],
                [_shift_toeplitz(m10), _shift_toeplitz(m11)],
            ],
        )

    shifted_transverse_permittivity_matrix = _shift_block_toeplitz(
        solve_result.transverse_permittivity_matrix
    )
    shifted_transverse_permeability_matrix = _shift_block_toeplitz(
        solve_result.transverse_permeability_matrix
    )

    shifted_z_permittivity_matrix = _shift_toeplitz(solve_result.z_permittivity_matrix)

    omega_script_k_matrix = fmm_matrices.omega_script_k_matrix_patterned(
        wavelength=solve_result.wavelength,
        z_permittivity_matrix=shifted_z_permittivity_matrix,
        transverse_permeability_matrix=shifted_transverse_permeability_matrix,
        transverse_wavevectors=basis.transverse_wavevectors(
            in_plane_wavevector=solve_result.in_plane_wavevector,
            primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
            expansion=solve_result.expansion,
        ),
    ).astype(solve_result.eigenvectors.dtype)

    return fmm.LayerSolveResult(
        wavelength=solve_result.wavelength,
        in_plane_wavevector=solve_result.in_plane_wavevector,
        primitive_lattice_vectors=solve_result.primitive_lattice_vectors,
        expansion=solve_result.expansion,
        eigenvalues=solve_result.eigenvalues,
        eigenvectors=shifted_eigenvectors,
        omega_script_k_matrix=omega_script_k_matrix,
        z_permittivity_matrix=shifted_z_permittivity_matrix,
        inverse_z_permittivity_matrix=_shift_toeplitz(
            solve_result.inverse_z_permittivity_matrix
        ),
        transverse_permittivity_matrix=shifted_transverse_permittivity_matrix,
        z_permeability_matrix=_shift_toeplitz(solve_result.z_permeability_matrix),
        inverse_z_permeability_matrix=_shift_toeplitz(
            solve_result.inverse_z_permeability_matrix
        ),
        transverse_permeability_matrix=shifted_transverse_permeability_matrix,
        tangent_vector_field=solve_result.tangent_vector_field,
    )


def _apply_shift_eigenvectors(
    eigenvectors: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    dx: jnp.ndarray,
    dy: jnp.ndarray,
) -> jnp.ndarray:
    """Returns eigenvectors that result from translating a structure."""
    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=jnp.zeros((2,)),
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    kx = transverse_wavevectors[:, 0]
    ky = transverse_wavevectors[:, 1]
    return eigenvectors * jnp.exp(-1j * (kx * dx + ky * dy))[..., jnp.newaxis]


def _apply_shift_toeplitz(
    mat: jnp.ndarray,
    expansion: basis.Expansion,
    primitive_lattice_vectors: basis.LatticeVectors,
    dx: jnp.ndarray,
    dy: jnp.ndarray,
) -> jnp.ndarray:
    """Return toeplitz matrix that results from translating a structure."""
    idx = fft._standard_toeplitz_indices(expansion)

    reciprocal_vectors = primitive_lattice_vectors.reciprocal
    u = reciprocal_vectors.u
    v = reciprocal_vectors.v
    kx = 2 * jnp.pi * (idx[:, :, 0] * u[0] + idx[:, :, 1] * v[0])
    ky = 2 * jnp.pi * (idx[:, :, 0] * u[1] + idx[:, :, 1] * v[1])
    return mat * jnp.exp(-1j * (kx * dx + ky * dy))



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Defines a jax-differentiable eigendecomposition.

Copyright (c) Martin F. Schubert
"""

from typing import Tuple

import jax
import jax.numpy as jnp

from fmmax import misc

# The `jeig` package offers several jax-wrapped implementations of eigendecomposition,
# some of which have performance benefits. However, since `jeig` has a dependency on
# pytorch, we make its use optional. If `jeig` is not available, we fall back on a
# pure-jax implementation of the eigendecomposition.
try:
    import jeig

    _JEIG_AVAILABLE = True
except ModuleNotFoundError:
    _JEIG_AVAILABLE = False


_EIG_EPS_RELATIVE = 1e-12
_EIG_EPS_MINIMUM = 1e-24


@jax.custom_vjp
def eig(
    matrix: jnp.ndarray,
    eps_relative: float = _EIG_EPS_RELATIVE,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Wraps `jnp.linalg.eig` in a jit-compatible, differentiable manner.

    The custom vjp allows gradients with resepct to the eigenvectors, unlike the
    standard jax implementation of `eig`. We use an expression for the gradient
    given in [2019 Boeddeker] along with a regularization scheme that applies
    a Lorentzian broadening to a term containing the inverse difference of
    eigenvalues. The broadening is related to the maximum magnitude of the
    eigenvalues.

    [2019 Boeddeker] https://arxiv.org/abs/1701.00392

    Args:
        matrix: The matrix for which eigenvalues and eigenvectors are sought.
        eps_relative: Parameter which determines the degree of broadening.

    Returns:
        The eigenvalues and eigenvectors.
    """
    del eps_relative
    return _eig(matrix)


def _eig_jax(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Eigendecomposition using `jax.lax.linalg.eig`."""
    eigenvalues, eigenvectors = jax.lax.linalg.eig(
        matrix,
        compute_left_eigenvectors=False,
        use_magma=False,
    )
    return eigenvalues, eigenvectors


def _eig(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Eigendecomposition using `jeig` if available, and `_eig_jax` if not."""
    if _JEIG_AVAILABLE:
        return jeig.eig(matrix)
    else:
        return _eig_jax(matrix)


def _eig_fwd(
    matrix: jnp.ndarray,
    eps_relative: float,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray, float]]:
    """Implements the forward calculation for `eig`."""
    eigenvalues, eigenvectors = _eig(matrix)
    return (eigenvalues, eigenvectors), (eigenvalues, eigenvectors, eps_relative)


def _eig_bwd(
    res: Tuple[jnp.ndarray, jnp.ndarray, float],
    grads: Tuple[jnp.ndarray, jnp.ndarray],
) -> Tuple[jnp.ndarray, None]:
    """Implements the backward calculation for `eig`."""
    eigenvalues, eigenvectors, eps_relative = res
    grad_eigenvalues, grad_eigenvectors = grads

    # The expression for gradient of matrix eigenvectors with respect to matrix values
    # contains a difference between eigenvalues in the denominator. This causes
    # numerical problems in situations where eigenvalues are degenerate or nearly
    # degenerate. This problem is addressed in a few different ways in existing rcwa
    # codes. Defining `delta = eigval_i - eigval_j`,
    #
    # - torcwa uses Lorentzian broadening of the form
    #      1 / delta -> delta.conj / (abs(delta)**2 + eps)
    #   with an eps value of 1e-10.
    #   (https://github.com/kch3782/torcwa/blob/main/torcwa/torch_eig.py#L29)
    #
    # - tf_rcwa uses an expression similar to torcwa, but with apparently different
    #   conjugation. Their value of eps is 1e-6. (We tested their implementation and
    #   found it fails some of our gradient validation tests.)
    #   (https://github.com/scolburn54/rcwa_tf/blob/master/src/tensor_utils.py#L89)
    #
    # - grcwa uses `1 / delta -> 1 / (delta + eps)` with an eps value of 1e-10.
    #   (https://github.com/weiliangjinca/grcwa/blob/master/grcwa/primitives.py#L44)
    #
    # Also, from the wider literature:
    #
    # - Liao et al. (https://arxiv.org/pdf/1903.09650 sec III.A.1) uses Lorentzian
    #   broadening with an eps value of 1e-12.
    #
    # One issue with Lorentzian broadening and a fixed eps is the independence from
    # matrix scale. While a given eps may be appropriate for a matrix, it would be
    # inappropriate for the same matrix scaled by e.g. 1e-6. (The gradient should
    # be identical, except scaled by 1 / 1e-6.)
    #
    # Therefore, we use Lorentzian broadening similar to torcwa, but with an eps
    # value that is computed in a way that considers the eigenvalue range.
    eigenvalues_i = eigenvalues[..., jnp.newaxis, :]
    eigenvalues_j = eigenvalues[..., :, jnp.newaxis]
    delta_eig = eigenvalues_i - eigenvalues_j
    eig_range_sq = jnp.amax(jnp.abs(delta_eig) ** 2, axis=(-2, -1), keepdims=True)
    eps = jnp.maximum(eps_relative * eig_range_sq, _EIG_EPS_MINIMUM)
    f_broadened = delta_eig.conj() / (jnp.abs(delta_eig) ** 2 + eps)

    # Manually set the diagonal elements to zero, as we do not use broadening here.
    i = jnp.arange(f_broadened.shape[-1])
    f_broadened = f_broadened.at[..., i, i].set(0)

    # By jax convention, gradients are with respect to the complex parameters, not with
    # respect to their conjugates. Take the conjugates.
    grad_eigenvalues_conj = jnp.conj(grad_eigenvalues)
    grad_eigenvectors_conj = jnp.conj(grad_eigenvectors)

    eigenvectors_adj = misc.matrix_adjoint(eigenvectors)
    dim = eigenvalues.shape[-1]
    eye_mask = jnp.eye(dim, dtype=bool)
    eye_mask = eye_mask.reshape((1,) * (eigenvalues.ndim - 1) + (dim, dim))

    # Then, the gradient is found by equation 4.77 of [2019 Boeddeker].
    rhs = (
        misc.diag(grad_eigenvalues_conj)
        + jnp.conj(f_broadened) * (eigenvectors_adj @ grad_eigenvectors_conj)
        - jnp.conj(f_broadened)
        * (eigenvectors_adj @ eigenvectors)
        @ jnp.where(eye_mask, jnp.real(eigenvectors_adj @ grad_eigenvectors_conj), 0.0)
    ) @ eigenvectors_adj
    grad_matrix = jnp.linalg.solve(eigenvectors_adj, rhs)

    # Take the conjugate of the gradient, reverting to the jax convention
    # where gradients are with respect to complex parameters.
    grad_matrix = jnp.conj(grad_matrix)

    # Return `grad_matrix`, and `None` for the gradient with respect to `eps`.
    return grad_matrix, None


eig.defvjp(_eig_fwd, _eig_bwd)



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Defines several utility functions.

Copyright (c) Martin F. Schubert
"""

from typing import Tuple

import jax
import jax.numpy as jnp


def solve(a: jnp.ndarray, b: jnp.ndarray, *, force_x64_solve: bool) -> jnp.ndarray:
    """Solves ``A @ x = b``, optionally using 64-bit precision."""
    output_dtype = jnp.promote_types(a.dtype, b.dtype)
    if force_x64_solve and jax.config.read("jax_enable_x64"):
        a = a.astype(jnp.promote_types(a.dtype, jnp.float64))
        b = b.astype(jnp.promote_types(b.dtype, jnp.float64))
    return jnp.linalg.solve(a, b).astype(output_dtype)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Following code is Copyright (c) Meta Platforms, Inc. and affiliates.
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


def interpolate_permittivity(
    permittivity_solid: jnp.ndarray,
    permittivity_void: jnp.ndarray,
    density: jnp.ndarray,
) -> jnp.ndarray:
    """Interpolates the permittivity with a scheme that avoids zero crossings.

    The interpolation uses the scheme introduced in [2019 Christiansen], which avoids
    zero crossings that can occur with metals or lossy materials having a negative
    real component of the permittivity. https://doi.org/10.1016/j.cma.2018.08.034

    Args:
        permittivity_solid: The permittivity of solid regions.
        permittivity_void: The permittivity of void regions.
        density: The density, specifying which locations are solid and which are void.

    Returns:
        The interpolated permittivity.
    """
    n_solid = jnp.real(jnp.sqrt(permittivity_solid))
    k_solid = jnp.imag(jnp.sqrt(permittivity_solid))
    n_void = jnp.real(jnp.sqrt(permittivity_void))
    k_void = jnp.imag(jnp.sqrt(permittivity_void))
    n = density * n_solid + (1 - density) * n_void
    k = density * k_solid + (1 - density) * k_void
    return (n + 1j * k) ** 2


def angular_frequency_for_wavelength(wavelength: jnp.ndarray) -> jnp.ndarray:
    """Returns the angular frequency for the specified wavelength."""
    return 2 * jnp.pi / wavelength  # Since by our convention c == 1.


def absolute_axes(axes: Tuple[int, ...], ndim: int) -> Tuple[int, ...]:
    """Returns the absolute axes for given relative axes and array dimensionality."""
    if not all(a in list(range(-ndim, ndim)) for a in axes):
        raise ValueError(
            f"All elements of `axes` must be in the range ({ndim}, {ndim - 1}) "
            f"but got {axes}."
        )
    absolute_axes = tuple([d % ndim for d in axes])
    if len(absolute_axes) != len(set(absolute_axes)):
        raise ValueError(
            f"Found duplicates in `axes`; computed absolute axes are {absolute_axes}."
        )
    return absolute_axes



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions related transforming to and from the Fourier basis.

Copyright (c) Martin F. Schubert
"""

from typing import Tuple

import jax.numpy as jnp

from fmmax import basis, utils


def fourier_convolution_matrix(
    x: jnp.ndarray,
    expansion: basis.Expansion,
) -> jnp.ndarray:
    """Computes the Fourier convolution matrix for ``x`` and ``basis_coefficients``.

    The Fourier convolution matrix at location ``(i, j)`` gives the Fourier
    coefficient associated with the lattice vector obtained by subtracting the
    ``j``th reciprocal lattice vector from the ``i``th reciprocal lattice basis.
    See equation 8 from [2012 Liu].

    Args:
        x: The array for which the Fourier coefficients are sought.
        expansion: The field expansion to be used.

    Returns:
        The coefficients, with shape ``(num_vectors, num_vectors)``.
    """
    basis.validate_shape_for_expansion(x.shape, expansion)

    x_fft = _fft2(x, axes=(-2, -1), norm="backward", centered_coordinates=True)
    x_fft /= jnp.prod(jnp.asarray(x.shape[-2:])).astype(x.dtype)
    idx = _standard_toeplitz_indices(expansion)
    return x_fft[..., idx[..., 0], idx[..., 1]]


def fft(
    x: jnp.ndarray,
    expansion: basis.Expansion,
    axes: Tuple[int, int] = (-2, -1),
    centered_coordinates: bool = True,
) -> jnp.ndarray:
    """Returns the 2D Fourier transform of ``x``.

    Args:
        x: The array to be transformed.
        expansion: The field expansion to be used.
        axes: The axes to be transformed, with default ``(-2, -1)``.
        centered_coordinates: Specifies the relationship between the physical
            coordinates and the indices of array elements in ``x``. If ``True``, the
            coordinates lie in the center of elements, so that the element with index
            ``(0, 0)`` lies at position ``du / 2 + dv / 2``. When ``False``, the
            coordinates lie in the corners of the elements, so that the ``(0, 0)``
            element lies at position ``0``.

    Returns:
        The transformed ``x``.
    """
    axes: Tuple[int, int] = utils.absolute_axes(axes, x.ndim)  # type: ignore[no-redef]
    basis.validate_shape_for_expansion(tuple([x.shape[ax] for ax in axes]), expansion)

    x_fft = _fft2(
        x,
        axes=axes,
        norm="forward",
        centered_coordinates=centered_coordinates,
    )

    leading_dims = len(x.shape[: axes[0]])
    trailing_dims = len(x.shape[axes[1] + 1 :])
    slices = (
        [slice(None)] * leading_dims
        + [expansion.basis_coefficients[:, 0], expansion.basis_coefficients[:, 1]]
        + [slice(None)] * trailing_dims
    )
    return x_fft[tuple(slices)]


def ifft(
    y: jnp.ndarray,
    expansion: basis.Expansion,
    shape: Tuple[int, int],
    axis: int = -1,
    centered_coordinates: bool = True,
) -> jnp.ndarray:
    """Returns the 2D inverse Fourier transform of ``y``.

    Args:
        y: The array to be transformed.
        expansion: The field expansion to be used.
        shape: The desired shape of the output array.
        axis: The axis containing the Fourier coefficients. Default is ``-1``, the
            final axis.
        centered_coordinates: Specifies the relationship between the physical
            coordinates and the indices of array elements in ``x``. If ``True``, the
            coordinates lie in the center of elements, so that the element with index
            ``(0, 0)`` lies at position ``du / 2 + dv / 2``. When ``False``, the
            coordinates lie in the corners of the elements, so that the ``(0, 0)``
            element lies at position ``0``.

    Returns:
        The inverse transformed ``y``.
    """
    (axis,) = utils.absolute_axes((axis,), y.ndim)
    assert y.shape[axis] == expansion.basis_coefficients.shape[-2]
    x_shape = y.shape[:axis] + shape + y.shape[axis + 1 :]
    assert len(x_shape) == y.ndim + 1

    basis.validate_shape_for_expansion(shape, expansion)

    leading_dims = len(y.shape[:axis])
    trailing_dims = len(y.shape[axis + 1 :])
    slices = (
        [slice(None)] * leading_dims
        + [expansion.basis_coefficients[:, 0], expansion.basis_coefficients[:, 1]]
        + [slice(None)] * trailing_dims
    )

    x = jnp.zeros(x_shape, y.dtype)
    x = x.at[tuple(slices)].set(y)
    return _ifft2(
        x,
        axes=(leading_dims, leading_dims + 1),
        norm="forward",
        centered_coordinates=centered_coordinates,
    )


def _fft2(
    x: jnp.ndarray,
    axes: Tuple[int, int] = (-2, -1),
    norm: str = "forward",
    centered_coordinates: bool = True,
) -> jnp.ndarray:
    """Two-dimensional Fourier transform."""
    y = jnp.fft.fft2(x, axes=axes, norm=norm)
    if centered_coordinates:
        axes = utils.absolute_axes(axes, ndim=x.ndim)  # type: ignore[assignment]
        ki = 0.5 * jnp.fft.fftfreq(x.shape[axes[0]])[:, jnp.newaxis]
        kj = 0.5 * jnp.fft.fftfreq(x.shape[axes[1]])[jnp.newaxis, :]
        phase = jnp.exp(-1j * 2 * jnp.pi * (ki + kj))
        phase = phase.reshape(phase.shape + (1,) * (x.ndim - axes[1] - 1))
        y = y * phase
    return y


def _ifft2(
    x: jnp.ndarray,
    axes: Tuple[int, int] = (-2, -1),
    norm: str = "forward",
    centered_coordinates: bool = True,
) -> jnp.ndarray:
    """Two-dimensional inverse Fourier transform."""
    if centered_coordinates:
        axes = utils.absolute_axes(axes, ndim=x.ndim)  # type: ignore[assignment]
        ki = 0.5 * jnp.fft.fftfreq(x.shape[axes[0]])[:, jnp.newaxis]
        kj = 0.5 * jnp.fft.fftfreq(x.shape[axes[1]])[jnp.newaxis, :]
        phase = jnp.exp(1j * 2 * jnp.pi * (ki + kj))
        phase = phase.reshape(phase.shape + (1,) * (x.ndim - axes[1] - 1))
        x = x * phase

    return jnp.fft.ifft2(x, axes=axes, norm=norm)


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Following code is Copyright (c) Meta Platforms, Inc. and affiliates.
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


def _standard_toeplitz_indices(expansion: basis.Expansion) -> jnp.ndarray:
    """Computes the indices for a standard Toeplitz matrix for `basis_coefficients`.

    Args:
        expansion: The field expansion to be used.

    Returns:
        The indices, with shape `(num, num, 2)`.
    """
    i, j = jnp.meshgrid(
        jnp.arange(expansion.num_terms),
        jnp.arange(expansion.num_terms),
        indexing="ij",
    )
    basis_coefficients = jnp.asarray(expansion.basis_coefficients)
    idx = basis_coefficients[i, :] - basis_coefficients[j, :]
    return idx



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Defines miscellaneous private functions.

Copyright (c) Martin F. Schubert
"""


from typing import Tuple

import jax.numpy as jnp

# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# Following code is Copyright (c) Meta Platforms, Inc. and affiliates.
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


def diag(x: jnp.ndarray) -> jnp.ndarray:
    """A batch-compatible version of `numpy.diag`."""
    shape = x.shape + (x.shape[-1],)
    y = jnp.zeros(shape, x.dtype)
    i = jnp.arange(x.shape[-1])
    return y.at[..., i, i].set(x)


def matrix_adjoint(x: jnp.ndarray) -> jnp.ndarray:
    """Computes the adjoint for a batch of matrices."""
    axes = tuple(range(x.ndim - 2)) + (x.ndim - 1, x.ndim - 2)
    return jnp.conj(jnp.transpose(x, axes=axes))


def batch_compatible_shapes(*shapes: Tuple[int, ...]) -> bool:
    """Returns `True` if all the shapes are batch-compatible."""
    max_dims = max([len(s) for s in shapes])
    shapes = tuple([(1,) * (max_dims - len(s)) + s for s in shapes])
    max_shape = [max(dim_shapes) for dim_shapes in zip(*shapes)]
    for shape in shapes:
        if any([a not in (1, b) for a, b in zip(shape, max_shape)]):
            return False
    return True


def atleast_nd(x: jnp.ndarray, n: int) -> jnp.ndarray:
    """Adds leading dimensions to `x`, ensuring that it is at least n-dimensional."""
    dims_to_add = tuple(range(max(0, n - x.ndim)))
    return jnp.expand_dims(x, axis=dims_to_add)
